//
// JTypist v3.1 - Java typing tutor program for UNIX systems
// Copyright (C) 2000  Simon Baldwin (simonb@sco.com)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//
// KBLayout class - defines a keyboard layout
//
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.text.MessageFormat;

/**
 * Defines a keyboard layout.
 */
public class KBLayout {

	// message catalog information
	private static ResourceBundle	messages	=
				ResourceBundle.getBundle( "MessageResource" );

	// the expected number of tokens for delete and replace operations
	private static final int	DEL_TOKENS		= 1;
	private static final int	REPL_TOKENS		= 5;

	// the radix of \\nnn character entries
	private static final int	ESC_RADIX		= 8;

	// description and keymap table; we use a vector since the ordering
	// of entries is used in ordering the statistics, and we want to
	// preserve it
	private String		kb_id;//		= null;
	private Vector		keymap_table		= new Vector();

	// lookup tables of glyphs by character they represent, and by keycode
	// they are assigned to, further looked up by shiftstate
	private Hashtable	glyphs_by_char		= new Hashtable();
	private Hashtable	glyphs_by_keycode_by_shiftstate
							= new Hashtable();


	/**
	 * Factory method to return a complete kblayout from the
	 * layout name given.
	 */
	public static KBLayout create_kblayout( String name ) {

		// try to find the resource bundle for the given name
		ResourceBundle resource		= null;
		try {
			resource = ResourceBundle.getBundle(
							"KBLayout" + name );
		}
		catch ( MissingResourceException e ) {}
		catch ( ClassCastException e ) {}
		if ( resource != null )
			return ( new KBLayout( resource ));

		// no resource bundle, so try for an old-style instantiation
		// of the class KBLayoutXXX.class
		try {
			Class class_handle = Class.forName( "KBLayout" + name );
			return ( (KBLayout)(class_handle.newInstance()) );
		} catch( Exception e ) {
			return ( null );
		}
	}


	/**
	 * Create an empty KBLayout - mostly used for fallback and recovery.
	 */
	public KBLayout() {}


	/**
	 * Create a new KBLayout from the given resource bundle.
	 */
	public KBLayout( ResourceBundle resource ) {

		// find any keyboard layout id, and store it
		try {
			this.kb_id = resource.getString( "id" );
		}
		catch ( MissingResourceException e ) {
			this.kb_id = messages.getString( "KBLayout.untitled" );
			String[] fmtargs = { kb_id };
			System.err.println( MessageFormat.format(
				messages.getString( "KBLayout.missingid" ),
								fmtargs ));
		}

		// find any includes spec, and deal with it
		try {
			String includes = resource.getString( "includes" );
			StringTokenizer tokenizer = new StringTokenizer(
								includes );
			while ( tokenizer.hasMoreTokens() ) {
				String include = tokenizer.nextToken();

				// try to get the kblayout for the include
				KBLayout kblayout =
					KBLayout.create_kblayout( include );
				if ( kblayout != null ) {

					// transfer the kblayout to ours
					for ( Enumeration enum = kblayout
							.get_elements();
						enum.hasMoreElements(); )
						replace( (Glyph)
							(enum.nextElement() ));
				}
				else {

					// include kblayout not found
					String[] fmtargs = { kb_id, include };
					System.err.println(
						MessageFormat.format(
							messages.getString(
						"KBLayout.badinclude" ),
								fmtargs ));
				}
			}
		}
		catch ( MissingResourceException e ) {}

		// find the index of entries
		String entry_index	= null;
		try {
			entry_index = resource.getString( "entry_index" );
		}
		catch ( MissingResourceException e ) {
			String[] fmtargs = { kb_id };
			System.err.println( MessageFormat.format(
					messages.getString(
						"KBLayout.missingindex" ),
								fmtargs ));
			return;
		}

		// handle each entry in the resource
		StringTokenizer entries	= new StringTokenizer( entry_index );
		while ( entries.hasMoreTokens() ) {
			String entry_ref = entries.nextToken();

			// look for this entry
			String entry	= null;
			try {
				entry = resource.getString( entry_ref );
			}
			catch ( MissingResourceException e ) {
				String[] fmtargs = { kb_id, entry_ref };
				System.err.println(
					MessageFormat.format(
						messages.getString(
						"KBLayout.missingentry" ),
								fmtargs ));
			}

			// handle this layout entry
			if ( entry != null )
				handle_entry( entry, entry_ref );
		}
	}

	/*
	 * Handle a resource entry string.
	 */
	private void handle_entry( String entry, String entry_ref ) {

		// error buffer for formatting messages
		Object[] fmtargs = { kb_id, entry_ref };

		// tokenize the entry on whitespace boundaries, and ensure
		// that it has either one or five tokens
		StringTokenizer tokenizer = new StringTokenizer( entry,
				messages.getString(
						"KBLayout.tokenseparator" ));
		int num_tokens		= tokenizer.countTokens();
		if ( num_tokens != DEL_TOKENS && num_tokens != REPL_TOKENS ) {
			System.err.println( MessageFormat.format(
				messages.getString(
						"KBLayout.wrongnumtokens" ),
				fmtargs ));
			return;
		}

		// get the character being represented; if it begins with
		// a backslash, this is an octal value for the character, and
		// if not, then it is the character itself (take just the
		// first character if there are more)
		String s_char		= tokenizer.nextToken();
		char c;
		if ( s_char.charAt( 0 ) == '\\' )
			try {
				c = (char)(Integer.parseInt(
					s_char.substring( 1 ), ESC_RADIX ));
			}
			catch ( NumberFormatException e ) {
				System.err.println( MessageFormat.format(
					messages.getString(
						"KBLayout.badoctal" ),
					fmtargs ));
				return;
			}
		else
			c = s_char.charAt( 0 );

		// if the line looks like a delete line, just delete this
		// character and we are done
		if ( num_tokens == DEL_TOKENS )
			this.delete( c );

		else {

			// it's a replace line, so get the rest of the fields
			String s_keycode	= tokenizer.nextToken();
			String s_shiftstate	= tokenizer.nextToken();
			String s_finger		= tokenizer.nextToken();
			String s_glyphgroup	= tokenizer.nextToken();

			// convert the keycode into the correct keycode object
			KeyCode keycode = KeyCode.find_keycode(
						s_keycode.toUpperCase() );
			if ( keycode == null ) {
				System.err.println( MessageFormat.format(
					messages.getString(
						"KBLayout.badkeycode" ),
					fmtargs ));
				return;
			}

			// get the correct shift state for the entry state
			KeyShiftState shiftstate
					= KeyShiftState.find_keyshiftstate(
						s_shiftstate.toUpperCase() );
			if ( shiftstate == null ) {
				System.err.println( MessageFormat.format(
					messages.getString(
						"KBLayout.badshiftstate" ),
					fmtargs ));
				return;
			}

			// get the finger that is represented in the entry
			Finger finger = Finger.find_finger(
						s_finger.toUpperCase() );
			if ( finger == null ) {
				System.err.println( MessageFormat.format(
					messages.getString(
						"KBLayout.badfinger" ),
					fmtargs ));
				return;
			}

			// get the glyph group that is represented by the entry
			GlyphGroup glyphgroup = GlyphGroup.find_glyphgroup(
						s_glyphgroup.toUpperCase() );
			if ( glyphgroup == null ) {
				System.err.println( MessageFormat.format(
					messages.getString(
						"KBLayout.badglyphgroup" ),
					fmtargs ));
				return;
			}

			// now that everything is validated, replace this entry
			// in the keyboard layout; if the character isn't
			// there already, this becomes an add operation
			this.replace( new Glyph(
				c, keycode, shiftstate, finger, glyphgroup ));
		}
	}


	/*
	 * Accessor methods.
	 */
	public String	get_kb_id()	{ return ( kb_id ); }
	public void	set_kb_id( String kb_id ) {
		this.kb_id = kb_id;
	}
	public Enumeration
			get_elements()	{ return ( keymap_table.elements() ); }


	/*
	 * Method to return the correct lookup table for a shiftstate.
	 */
	private Hashtable get_table( KeyShiftState shiftstate ) {

		// see if we already have a table for this shiftstate
		Hashtable hashtable = (Hashtable)
			(glyphs_by_keycode_by_shiftstate.get( shiftstate ));
		if ( hashtable == null ) {

			// not seen this shiftstate before, so create a new
			// lookup table for it
			hashtable = new Hashtable();
			glyphs_by_keycode_by_shiftstate.put(
							shiftstate, hashtable );
		}

		// return the right table for this shiftstate
		return ( hashtable );
	}

	/**
	 * Method to add a keymap entry into the table.  This places the new
	 * glyph at the end of the current table.
	 */
	public void append( Glyph glyph ) {

		// add the entry to the main list
		keymap_table.addElement( glyph );

		// add the entry to the lookup tables
		glyphs_by_char.put( glyph.get_character(), glyph );
		Hashtable hashtable = get_table( glyph.get_shiftstate() );
		hashtable.put( glyph.get_keycode(), glyph );
	}

	/**
	 * Method to replace an existing keymap entry in the table with
	 * another, retaining the position in the list of the original.
	 * The existing entry is a glyph having the same character.  If
	 * none exists, the glyph is added to the table.
	 */
	public void replace( Glyph glyph ) {

		// get the character for this glyph
		Character glyph_character	= glyph.get_character();

		// now search for any existing glyphs that implement the same
		// keyboard character
		Glyph old_glyph = (Glyph)
					(glyphs_by_char.get( glyph_character ));
		if ( old_glyph != null ) {

			// found one, so replace it with the new one, at the
			// same point in the original list
			int index = keymap_table.indexOf( old_glyph );
			keymap_table.setElementAt( glyph, index );

			// update the lookup tables for the replaced glyph
			glyphs_by_char.put( glyph.get_character(), glyph );
			Hashtable hashtable = get_table(
						glyph.get_shiftstate() );
			hashtable.put( glyph.get_keycode(), glyph );

			// since we found and replaced the glyph, we can
			// stop here
			return;
		}

		// there was nothing to replace, so instead the new glyph
		// is simply appended to the list
		append( glyph );
	}

	/**
	 * Method to delete a glyph from the keymap table.
	 */
	public void delete( char character ) {

		// search for any existing glyphs that implement the character
		Glyph old_glyph = (Glyph)(glyphs_by_char.get(
						new Character( character )));
		if ( old_glyph != null ) {

			// found a match, so remove it from the table
			keymap_table.removeElement( old_glyph );

			// remove the matched glyph from the lookup tables too
			glyphs_by_char.remove( old_glyph.get_character() );
			Hashtable hashtable = get_table(
						old_glyph.get_shiftstate() );
			hashtable.remove( old_glyph.get_keycode() );
		}
	}


	/**
	 * Find the glyph that represents a given character.  This returns
	 * a glyph, or null if no matching glyph exists.
	 */
	public Glyph find_glyph( char character ) {

		// return a simple lookup
		return ( (Glyph)(glyphs_by_char.get(
						new Character( character ))) );
	}

	/**
	 * Find the glyph that represents a given keycode, with shift state.
	 * This returns a glyph, or null if no matching glyph exists.
	 */
	public Glyph find_glyph( KeyCode keycode, KeyShiftState shiftstate ) {

		// return a simple lookup
		Hashtable hashtable = get_table( shiftstate );
		return ( (Glyph)(hashtable.get( keycode )) );
	}


	/**
	 * Printout method.
	 */
	public String toString() {
		return ( getClass().getName() + "{"
				+ "kb_id=\"" + kb_id
				+ "\",keymap_table@" + keymap_table
				+ "}" );
	}


	/**
	 * Class test code.
	 */
	public static void main ( String args[] ) {
		System.out.println( "*** CLASS TEST CODE ***" );

		System.out.println( KBLayout.create_kblayout( "US" ));
	}
}
